#  Copyright (c) 2016 IBM Corporation
#  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import collections
import copy
import distutils.version as dist_version  # pylint: disable=E0611
import math
import os
import string

import eventlet
from oslo_log import log as logging
import six

from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder.objects import fields
import cinder.volume.drivers.ibm.ibm_storage as storage
from cinder.volume.drivers.ibm.ibm_storage import cryptish
from cinder.volume.drivers.ibm.ibm_storage import ds8k_restclient as restclient
from cinder.volume.drivers.ibm.ibm_storage import proxy
from cinder.volume.drivers.ibm.ibm_storage import strings

LOG = logging.getLogger(__name__)

LSS_VOL_SLOTS = 0x100
LSS_SLOTS = 0xFF

VALID_HOST_TYPES = (
    'auto', 'AMDLinuxRHEL', 'AMDLinuxSuse',
    'AppleOSX', 'Fujitsu', 'Hp', 'HpTru64',
    'HpVms', 'LinuxDT', 'LinuxRF', 'LinuxRHEL',
    'LinuxSuse', 'Novell', 'SGI', 'SVC',
    'SanFsAIX', 'SanFsLinux', 'Sun', 'VMWare',
    'Win2000', 'Win2003', 'Win2008', 'Win2012',
    'iLinux', 'nSeries', 'pLinux', 'pSeries',
    'pSeriesPowerswap', 'zLinux', 'iSeries'
)


def filter_alnum(s):
    return ''.join(x if x in string.ascii_letters +
                   string.digits else '_' for x in s) if s else ''


class DS8KCommonHelper(object):
    """Manage the primary backend, it is common class too."""

    OPTIONAL_PARAMS = ['ds8k_host_type', 'lss_range_for_cg']
    # if use new REST API, please update the version below
    VALID_REST_VERSION_87_51_MIN = '87.51.52.0'
    INVALID_STORAGE_VERSION = '8.0.1'
    REST_VERSION_87_51_MIN_PPRC_CG = '87.51.63.0'
    REST_VERSION_88_20_MIN_PPRC_CG = '88.20.112.0'

    def __init__(self, conf, HTTPConnectorObject=None):
        self.conf = conf
        self._connector_obj = HTTPConnectorObject
        self._storage_pools = None
        self._disable_thin_provision = False
        self._connection_type = self._get_value('connection_type')
        self._existing_lss = None
        self.backend = {}
        self.setup()

    @staticmethod
    def _gb2b(gb):
        return gb * (2 ** 30)

    def _get_value(self, key):
        if getattr(self.conf, 'safe_get', 'get') == 'get':
            value = self.conf.get(key)
        else:
            value = self.conf.safe_get(key)
        if not value and key not in self.OPTIONAL_PARAMS:
            raise exception.InvalidParameterValue(
                err=(_('Param [%s] should be provided.') % key))
        return value

    def get_thin_provision(self):
        return self._disable_thin_provision

    def get_storage_pools(self):
        return self._storage_pools

    def get_connection_type(self):
        return self._connection_type

    def get_pool(self, lss):
        node = int(lss, 16) % 2
        pids = [
            pid for pid, p in self._storage_pools.items() if p['node'] == node]
        return pids[0] if pids else None

    def setup(self):
        self._create_client()
        self._get_storage_information()
        self._check_host_type()
        self.backend['pools_str'] = self._get_value('san_clustername')
        self._storage_pools = self.get_pools()
        self.verify_pools(self._storage_pools)
        self.backend['lss_ids_for_cg'] = self._get_lss_ids_for_cg()
        self._verify_rest_version()

    def update_client(self):
        self._client.close()
        self._create_client()

    def _get_certificate(self, host):
        cert_file = strings.CERTIFICATES_PATH + host + '.pem'
        LOG.debug("certificate file for DS8K %(host)s: %(cert)s",
                  {'host': host, 'cert': cert_file})
        # Use the certificate if it exists, otherwise use the System CA Bundle
        if os.path.exists(cert_file):
            return cert_file
        else:
            LOG.debug("certificate file not found.")
            return True

    def _create_client(self):
        san_ip = self._get_value('san_ip')
        try:
            clear_pass = cryptish.decrypt(self._get_value('san_password'))
        except TypeError:
            raise exception.InvalidParameterValue(
                err=_('Param [san_password] is invalid.'))
        verify = self._get_certificate(san_ip)
        try:
            self._client = restclient.RESTScheduler(
                san_ip,
                self._get_value('san_login'),
                clear_pass,
                self._connector_obj,
                verify)
        except restclient.TimeoutException:
            raise restclient.APIException(
                data=(_("Can't connect to %(host)s") % {'host': san_ip}))
        self.backend['rest_version'] = self._get_version()['bundle']
        LOG.info("Connection to DS8K storage system %(host)s has been "
                 "established successfully, the version of REST is %(rest)s.",
                 {'host': self._get_value('san_ip'),
                  'rest': self.backend['rest_version']})

    def _get_storage_information(self):
        storage_info = self.get_systems()
        self.backend['storage_unit'] = storage_info['id']
        self.backend['storage_wwnn'] = storage_info['wwnn']
        self.backend['storage_version'] = storage_info['release']

    def _get_lss_ids_for_cg(self):
        lss_range = self._get_value('lss_range_for_cg')
        lss_ids_for_cg = set()
        if not lss_range or not lss_range.strip():
            return lss_ids_for_cg
        if '-' in lss_range:
            lss_ids = list()
            lss_range = lss_range.strip()
            if lss_range.startswith('-') or lss_range.endswith('-'):
                raise exception.InvalidParameterValue(
                    err=_('Param [lss_range_for_cg]\'s format is invalid, '
                          'please don\'t put the \'-\' at the beginning or '
                          'the end.'))
            lss_range = lss_range.replace('-', ' - ').split()
            for index, lss in enumerate(lss_range):
                if lss == '-':
                    try:
                        begin = int(lss_range[index - 1], 16)
                        end = int(lss_range[index + 1], 16)
                        lss_ids_for_cg |= set(
                            ('%02x' % i).upper() for i in
                            range(begin, end + 1))
                    except ValueError as e:
                        raise exception.InvalidParameterValue(
                            err=_('Param [lss_range_for_cg] is invalid, it '
                                  'only supports space and \'-\' as '
                                  'separator. '
                                  'Exception = %s.') % six.text_type(e))
                else:
                    lss_ids.append(lss)
            lss_ids_for_cg |= set(lss_ids)
        else:
            lss_ids_for_cg = set(lss_range.split())
        for lss_id in lss_ids_for_cg:
            try:
                if int(lss_id, 16) > 0xFF:
                    raise exception.InvalidParameterValue(
                        err=_('Param [lss_range_for_cg] is invalid, it '
                              'should be within 00-FF'))
            except ValueError as e:
                raise exception.InvalidParameterValue(
                    err=_('Param [lss_range_for_cg] is invalid, it '
                          'only supports space and \'-\' as separator. '
                          'Exception = %s.') % six.text_type(e))
        return lss_ids_for_cg

    def _check_host_type(self):
        ds8k_host_type = self._get_value('ds8k_host_type')
        if (ds8k_host_type and
           (ds8k_host_type not in VALID_HOST_TYPES)):
            msg = (_("Param [ds8k_host_type] must be one of: %(values)s.")
                   % {'values': VALID_HOST_TYPES[1:-1]})
            LOG.error(msg)
            raise exception.InvalidParameterValue(err=msg)
        self.backend['host_type_override'] = (
            None if ds8k_host_type == 'auto' else ds8k_host_type)

    def _verify_rest_version(self):
        if self.backend['storage_version'] == self.INVALID_STORAGE_VERSION:
            raise exception.VolumeDriverException(
                message=(_("%s does not support bulk deletion of volumes, "
                           "if you want to use this version of driver, "
                           "please upgrade the CCL.")
                         % self.INVALID_STORAGE_VERSION))
        rest_ver = self.backend['rest_version'][0:2]
        if ('87' == rest_ver and
           dist_version.LooseVersion(self.backend['rest_version']) <
           dist_version.LooseVersion(self.VALID_REST_VERSION_87_51_MIN)):
            raise exception.VolumeDriverException(
                message=(_("REST version %(invalid)s is lower than "
                           "%(valid)s, please upgrade it in DS8K.")
                         % {'invalid': self.backend['rest_version'],
                            'valid': self.VALID_REST_VERSION_87_51_MIN}))

    def verify_rest_version_for_pprc_cg(self):
        if '8.1' in self.backend['rest_version']:
            raise exception.VolumeDriverException(
                message=_("REST for DS8K 8.1 does not support PPRC "
                          "consistency group, please upgrade the CCL."))
        valid_rest_version = None
        rest_ver = self.backend['rest_version'][0:2]
        if ('87' == rest_ver and
           dist_version.LooseVersion(self.backend['rest_version']) <
           dist_version.LooseVersion(self.REST_VERSION_87_51_MIN_PPRC_CG)):
            valid_rest_version = self.REST_VERSION_87_51_MIN_PPRC_CG
        elif ('88' == rest_ver and
              dist_version.LooseVersion(self.backend['rest_version']) <
              dist_version.LooseVersion(self.REST_VERSION_88_20_MIN_PPRC_CG)):
            valid_rest_version = self.REST_VERSION_88_20_MIN_PPRC_CG

        if valid_rest_version:
            raise exception.VolumeDriverException(
                message=(_("REST version %(invalid)s is lower than "
                           "%(valid)s, please upgrade it in DS8K.")
                         % {'invalid': self.backend['rest_version'],
                            'valid': valid_rest_version}))

    def verify_pools(self, storage_pools):
        if self._connection_type == storage.XIV_CONNECTION_TYPE_FC:
            ptype = 'fb'
        elif self._connection_type == storage.XIV_CONNECTION_TYPE_FC_ECKD:
            ptype = 'ckd'
        else:
            raise exception.InvalidParameterValue(
                err=_('Param [connection_type] is invalid.'))
        for pid, pool in storage_pools.items():
            if pool['stgtype'] != ptype:
                LOG.error('The stgtype of pool %(pool)s is %(ptype)s.',
                          {'pool': pid, 'ptype': pool['stgtype']})
                raise exception.InvalidParameterValue(
                    err='Param [san_clustername] is invalid.')

    @proxy.logger
    def get_pools(self, specific_pools=None):
        if specific_pools:
            pools_str = specific_pools.replace(' ', '').upper().split(',')
        else:
            pools_str = self.backend['pools_str'].replace(
                ' ', '').upper().split(',')
        pools = []
        storage_pools = collections.OrderedDict()
        for pid in pools_str:
            try:
                pools.append(self._get_pool(pid))
            except restclient.APIException as e:
                LOG.warning("Failed to get pool %(id)s information, "
                            "Exception: %(ex)s.", {'id': pid,
                                                   'ex': six.text_type(e)})
        if len(pools):
            unsorted_pools = self._format_pools(pools)
            storage_pools = collections.OrderedDict(sorted(
                unsorted_pools, key=lambda i: i[1]['capavail'], reverse=True))
        return storage_pools

    @proxy.logger
    def update_storage_pools(self, storage_pools):
        self._storage_pools = storage_pools

    def _format_pools(self, pools):
        return ((p['id'], {
            'name': p['name'],
            'node': int(p['node']),
            'stgtype': p['stgtype'],
            'cap': int(p['cap']),
            'capavail': int(p['capavail'])
        }) for p in pools)

    def verify_lss_ids(self, specified_lss_ids):
        if not specified_lss_ids:
            return None
        lss_ids = specified_lss_ids.upper().replace(' ', '').split(',')
        # verify LSS IDs.
        for lss_id in lss_ids:
            if int(lss_id, 16) > 255:
                raise exception.InvalidParameterValue(
                    _('LSS %s should be within 00-FF.') % lss_id)
        # verify address group
        self._existing_lss = self.get_all_lss()
        ckd_addrgrps = set(int(lss['id'], 16) // 16 for lss in
                           self._existing_lss if lss['type'] == 'ckd')
        fb_addrgrps = set((int(lss, 16) // 16) for lss in lss_ids)
        intersection = ckd_addrgrps & fb_addrgrps
        if intersection:
            raise exception.VolumeDriverException(
                message=_('LSSes in the address group %s are reserved '
                          'for CKD volumes') % list(intersection))
        # verify whether LSSs specified have been reserved for
        # consistency group or not.
        if self.backend['lss_ids_for_cg']:
            for lss_id in lss_ids:
                if lss_id in self.backend['lss_ids_for_cg']:
                    raise exception.InvalidParameterValue(
                        _('LSS %s has been reserved for CG.') % lss_id)
        return lss_ids

    @proxy.logger
    def find_pool_lss_pair(self, pool, find_new_pid, excluded_lss):
        if pool:
            node = int(pool[1:], 16) % 2
            lss = self._find_lss(node, excluded_lss)
            if lss:
                return (pool, lss)
            else:
                if not find_new_pid:
                    raise restclient.LssIDExhaustError(
                        message=_('All LSS/LCU IDs for configured pools '
                                  'on storage are exhausted.'))
        # find new pool id and lss for lun
        return self.find_biggest_pool_and_lss(excluded_lss)

    @proxy.logger
    def find_biggest_pool_and_lss(self, excluded_lss, specified_pool_lss=None):
        if specified_pool_lss:
            # pool and lss should be verified every time user create volume or
            # snapshot, because they can be changed in extra-sepcs at any time.
            specified_pool_ids, specified_lss_ids = specified_pool_lss
            storage_pools = self.get_pools(specified_pool_ids)
            self.verify_pools(storage_pools)
            storage_lss = self.verify_lss_ids(specified_lss_ids)
        else:
            storage_pools, storage_lss = self._storage_pools, None
        # pools are ordered by capacity
        for pool_id, pool in storage_pools.items():
            lss = self._find_lss(pool['node'], excluded_lss, storage_lss)
            if lss:
                return pool_id, lss
        raise restclient.LssIDExhaustError(
            message=_("All LSS/LCU IDs for configured pools are exhausted."))

    @proxy.logger
    def _find_lss(self, node, excluded_lss, specified_lss_ids=None):
        if specified_lss_ids:
            existing_lss = self._existing_lss
        else:
            existing_lss = self.get_all_lss()
        LOG.info("Existing LSS IDs are: %s.",
                 ','.join([lss['id'] for lss in existing_lss]))
        saved_existing_lss = copy.copy(existing_lss)

        # exclude LSSs that are full.
        existing_lss = [lss for lss in existing_lss
                        if lss['id'] not in excluded_lss]
        if not existing_lss:
            LOG.info("All LSSs are full.")
            return None

        # user specify LSSs in extra-specs.
        if specified_lss_ids:
            specified_lss_ids = [lss for lss in specified_lss_ids
                                 if lss not in excluded_lss]
            if specified_lss_ids:
                existing_lss = [lss for lss in existing_lss
                                if lss['id'] in specified_lss_ids]
                nonexistent_lss_ids = (set(specified_lss_ids) -
                                       set(lss['id'] for lss in existing_lss))
                lss = None
                for lss_id in nonexistent_lss_ids:
                    if int(lss_id, 16) % 2 == node:
                        lss = lss_id
                        break
                if not lss:
                    lss = self._find_from_existing_lss(
                        node, existing_lss, True)
            else:
                LOG.info("All appropriate LSSs specified are full.")
                return None
        else:
            # exclude LSSs that reserved for CG.
            if self.backend['lss_ids_for_cg']:
                existing_lss_cg, nonexistent_lss_cg = (
                    self._classify_lss_for_cg(existing_lss))
                existing_lss = [lss for lss in existing_lss
                                if lss['id'] not in existing_lss_cg]
            else:
                existing_lss_cg = set()
                nonexistent_lss_cg = set()
            lss = self._find_from_existing_lss(node, existing_lss)
            if not lss:
                lss = self._find_from_nonexistent_lss(node, saved_existing_lss,
                                                      nonexistent_lss_cg)
        return lss

    def _classify_lss_for_cg(self, existing_lss):
        existing_lss_ids = set(lss['id'] for lss in existing_lss)
        existing_lss_cg = existing_lss_ids & self.backend['lss_ids_for_cg']
        nonexistent_lss_cg = self.backend['lss_ids_for_cg'] - existing_lss_cg
        return existing_lss_cg, nonexistent_lss_cg

    def _find_from_existing_lss(self, node, existing_lss, ignore_pprc=False):
        if not ignore_pprc:
            # exclude LSSs that are used by PPRC paths.
            lss_in_pprc = self.get_lss_in_pprc_paths()
            if lss_in_pprc:
                existing_lss = [lss for lss in existing_lss
                                if lss['id'] not in lss_in_pprc]
        # exclude wrong type of LSSs and those that are not in expected node.
        existing_lss = [lss for lss in existing_lss if lss['type'] == 'fb'
                        and int(lss['group']) == node]
        lss_id = None
        if existing_lss:
            # look for the emptiest lss from existing lss
            lss = sorted(existing_lss, key=lambda k: int(k['configvols']))[0]
            if int(lss['configvols']) < LSS_VOL_SLOTS:
                lss_id = lss['id']
                LOG.info('_find_from_existing_lss: choose %(lss)s. '
                         'now it has %(num)s volumes.',
                         {'lss': lss_id, 'num': lss['configvols']})
        return lss_id

    def _find_from_nonexistent_lss(self, node, existing_lss, lss_cg=None):
        ckd_addrgrps = set(int(lss['id'], 16) // 16 for lss in existing_lss if
                           lss['type'] == 'ckd' and int(lss['group']) == node)
        full_lss = set(int(lss['id'], 16) for lss in existing_lss if
                       lss['type'] == 'fb' and int(lss['group']) == node)
        cg_lss = set(int(lss, 16) for lss in lss_cg) if lss_cg else set()
        # look for an available lss from nonexistent lss
        lss_id = None
        for lss in range(node, LSS_SLOTS, 2):
            addrgrp = lss // 16
            if (addrgrp not in ckd_addrgrps and
               lss not in full_lss and
               lss not in cg_lss):
                lss_id = ("%02x" % lss).upper()
                break
        LOG.info('_find_from_unexisting_lss: choose %s.', lss_id)
        return lss_id

    def create_lun(self, lun):
        volData = {
            'cap': self._gb2b(lun.size),
            'captype': 'bytes',
            'stgtype': 'fb',
            'tp': 'ese' if lun.type_thin else 'none'
        }
        lun.data_type = lun.data_type if lun.data_type else 'FB 512'
        if lun.type_os400:
            volData['os400'] = lun.type_os400
        volData['name'] = lun.ds_name
        volData['pool'], volData['lss'] = lun.pool_lss_pair['source']
        lun.ds_id = self._create_lun(volData)
        return lun

    def delete_lun(self, luns):
        lun_ids = []
        luns = [luns] if not isinstance(luns, list) else luns
        for lun in luns:
            if lun.ds_id is None:
                # create_lun must have failed and not returned the id
                LOG.error("delete_lun: volume id is None.")
                continue
            if not self.lun_exists(lun.ds_id):
                LOG.error("delete_lun: volume %s not found.", lun.ds_id)
                continue
            lun_ids.append(lun.ds_id)

        # Max 32 volumes could be deleted by specifying ids parameter
        while lun_ids:
            if len(lun_ids) > 32:
                lun_ids_str = ','.join(lun_ids[0:32])
                del lun_ids[0:32]
            else:
                lun_ids_str = ','.join(lun_ids)
                lun_ids = []
            LOG.info("Deleting volumes: %s.", lun_ids_str)
            self._delete_lun(lun_ids_str)

    def get_lss_in_pprc_paths(self):
        # TODO(Jiamin): when the REST API that get the licenses installed
        # in DS8K is ready, this function should be improved.
        try:
            paths = self.get_pprc_paths()
        except restclient.APIException:
            paths = []
            LOG.exception("Can not get the LSS")
        lss_ids = set(p['source_lss_id'] for p in paths)
        LOG.info('LSS in PPRC paths are: %s.', ','.join(lss_ids))
        return lss_ids

    def wait_flashcopy_finished(self, src_luns, tgt_luns):
        valid_fc_states = ('valid', 'validation_required')
        for tgt_lun in tgt_luns:
            tgt_lun.status = 'checking'
        while True:
            eventlet.sleep(5)
            for src_lun, tgt_lun in zip(src_luns, tgt_luns):
                if tgt_lun.status == 'checking':
                    try:
                        fcs = self.get_flashcopy(tgt_lun.ds_id)
                        if not fcs:
                            tgt_lun.status = 'available'
                        elif fcs[0]['state'] not in valid_fc_states:
                            LOG.error('Flashcopy %(src)s:%(tgt)s ended '
                                      'up in bad state %(state)s.',
                                      {'src': src_lun.ds_id,
                                       'tgt': tgt_lun.ds_id,
                                       'state': fcs[0]['state']})
                            tgt_lun.status = 'error'
                    except restclient.APIException:
                        LOG.error('Can not get flashcopy relationship '
                                  '%(src)s:%(tgt)s',
                                  {'src': src_lun.ds_id,
                                   'tgt': tgt_lun.ds_id})
                        tgt_lun.status = 'error'
            if not [lun for lun in tgt_luns if lun.status == 'checking']:
                break
        # cleanup error flashcopy relationship.
        for src_lun, tgt_lun in zip(src_luns, tgt_luns):
            if tgt_lun.status == 'error':
                self.delete_flashcopy(src_lun.ds_id, tgt_lun.ds_id)

    def wait_pprc_copy_finished(self, vol_ids, state, delete=True):
        LOG.info("Wait for PPRC pair to enter into state %s", state)
        vol_ids = sorted(vol_ids)
        min_vol_id = min(vol_ids)
        max_vol_id = max(vol_ids)
        invalid_states = ('target_suspended',
                          'invalid',
                          'volume_inaccessible')
        if state == 'full_duplex':
            invalid_states += ('suspended',)
        elif state == 'suspended':
            invalid_states += ('valid',)
        try:
            finished = False
            while True:
                eventlet.sleep(2)
                pairs = self.get_pprc_pairs(min_vol_id, max_vol_id)
                pairs = [
                    p for p in pairs if p['source_volume']['name'] in vol_ids]
                finished_pairs = [p for p in pairs if p['state'] == state]
                if len(finished_pairs) == len(pairs):
                    finished = True
                    break
                unfinished_pairs = [p for p in pairs if p['state'] != state]
                for p in unfinished_pairs:
                    if p['state'] in invalid_states:
                        raise restclient.APIException(
                            data=(_('Metro Mirror pair %(id)s enters into '
                                    'state %(state)s. ')
                                  % {'id': p['id'], 'state': p['state']}))
        finally:
            if not finished and delete:
                pair_ids = {'ids': ','.join([p['id'] for p in pairs])}
                self.delete_pprc_pair_by_pair_id(pair_ids)

    def _get_host(self, connector):
        # DS8K doesn't support hostname which is longer than 32 chars.
        hname = filter_alnum(connector['host'])[:32]
        os_type = connector.get('os_type')
        platform = connector.get('platform')

        if self.backend['host_type_override']:
            htype = self.backend['host_type_override']
        elif os_type == 'OS400':
            htype = 'iSeries'
        elif os_type == 'AIX':
            htype = 'pSeries'
        elif platform in ('s390', 's390x') and os_type in ('linux', 'linux2'):
            htype = 'zLinux'
        else:
            htype = 'LinuxRHEL'
        return collections.namedtuple('Host', ('name', 'type'))(hname, htype)

    def check_vol_mapped_to_host(self, connector, vol_id):
        map_info = {
            'host_ports': [],
            'mappings': [],
            'lun_ids': []
        }
        host_wwpn_set = set(wwpn.upper() for wwpn in connector['wwpns'])
        host_ports = self._get_host_ports(host_wwpn_set)
        defined_hosts = set(
            hp['host_id'] for hp in host_ports if hp['host_id'])
        if not defined_hosts:
            return False, None, map_info
        elif len(defined_hosts) > 1:
            raise restclient.APIException(_('More than one host found.'))
        else:
            host_id = defined_hosts.pop()
            mappings = self._get_mappings(host_id)
            lun_ids = [
                m['lunid'] for m in mappings if m['volume']['id'] == vol_id]
            map_info['host_ports'] = host_ports
            map_info['mappings'] = mappings
            map_info['lun_ids'] = lun_ids
            if not lun_ids:
                return False, host_id, map_info
            else:
                return True, host_id, map_info

    @coordination.synchronized('ibm-ds8k-{connector[host]}')
    def initialize_connection(self, vol_id, connector, **kwargs):
        host = self._get_host(connector)
        # Find defined host and undefined host ports
        host_wwpn_set = set(wwpn.upper() for wwpn in connector['wwpns'])
        host_ports = self._get_host_ports(host_wwpn_set)
        LOG.debug("host_ports: %s", host_ports)
        defined_hosts = set(
            hp['host_id'] for hp in host_ports if hp['host_id'])
        unknown_ports = host_wwpn_set - set(
            hp['wwpn'] for hp in host_ports)
        unconfigured_ports = set(
            hp['wwpn'] for hp in host_ports if not hp['host_id'])
        LOG.debug("initialize_connection: defined_hosts: %(defined)s, "
                  "unknown_ports: %(unknown)s, unconfigured_ports: "
                  "%(unconfigured)s.", {"defined": defined_hosts,
                                        "unknown": unknown_ports,
                                        "unconfigured": unconfigured_ports})
        # Create host if it is not defined
        if not defined_hosts:
            host_id = self._create_host(host)['id']
        elif len(defined_hosts) == 1:
            host_id = defined_hosts.pop()
        else:
            raise restclient.APIException(
                message='More than one host defined for requested ports.')
        LOG.info('Volume will be attached to host %s.', host_id)

        # Create missing host ports
        if unknown_ports or unconfigured_ports:
            self._assign_host_port(host_id,
                                   list(unknown_ports | unconfigured_ports))

        # Map the volume to host
        lun_id = self._map_volume_to_host(host_id, vol_id)
        target_ports = [p['wwpn'] for p in self._get_ioports()]
        return {
            'driver_volume_type': 'fibre_channel',
            'data': {
                'target_discovered': False,
                'target_lun': int(lun_id, 16),
                'target_wwn': target_ports,
                'initiator_target_map': {initiator: target_ports for
                                         initiator in connector['wwpns']}
            }
        }

    @coordination.synchronized('ibm-ds8k-{connector[host]}')
    def terminate_connection(self, vol_id, host_id, connector, map_info):
        host = self._get_host(connector)
        host_ports = map_info['host_ports']
        lun_ids = map_info['lun_ids']
        mappings = map_info['mappings']
        delete_ports = set(
            hp['wwpn'] for hp in host_ports if not hp['host_id'])
        LOG.debug("terminate_connection: host_ports: %(host)s, "
                  "defined_hosts: %(defined)s, delete_ports: %(delete)s.",
                  {"host": host_ports,
                   "defined": host_id,
                   "delete": delete_ports})
        for lun_id in lun_ids:
            self._delete_mappings(host_id, lun_id)
        if not lun_ids:
            LOG.warning("Volume %(vol)s is already not mapped to "
                        "host %(host)s.",
                        {'vol': vol_id, 'host': host.name})
        # if this host only has volumes that have been detached,
        # remove the host and its ports
        ret_info = {
            'driver_volume_type': 'fibre_channel',
            'data': {}
        }
        if len(mappings) == len(lun_ids):
            for port in delete_ports:
                self._delete_host_ports(port)
            self._delete_host(host_id)
            target_ports = [p['wwpn'] for p in self._get_ioports()]
            target_map = {initiator: target_ports
                          for initiator in connector['wwpns']}
            ret_info['data']['initiator_target_map'] = target_map
        return ret_info

    def create_group(self, group):
        return {'status': fields.GroupStatus.AVAILABLE}

    def delete_group(self, group, src_luns):
        volumes_model_update = []
        model_update = {'status': fields.GroupStatus.DELETED}
        if src_luns:
            try:
                self.delete_lun(src_luns)
            except restclient.APIException as e:
                model_update['status'] = fields.GroupStatus.ERROR_DELETING
                LOG.exception(
                    "Failed to delete the volumes in group %(group)s, "
                    "Exception = %(ex)s",
                    {'group': group.id, 'ex': e})

            for src_lun in src_luns:
                volumes_model_update.append({
                    'id': src_lun.os_id,
                    'status': model_update['status']
                })
        return model_update, volumes_model_update

    def delete_group_snapshot(self, group_snapshot, tgt_luns):
        snapshots_model_update = []
        model_update = {'status': fields.GroupSnapshotStatus.DELETED}
        if tgt_luns:
            try:
                self.delete_lun(tgt_luns)
            except restclient.APIException as e:
                model_update['status'] = (
                    fields.GroupSnapshotStatus.ERROR_DELETING)
                LOG.error("Failed to delete snapshots in group snapshot "
                          "%(gsnapshot)s, Exception = %(ex)s",
                          {'gsnapshot': group_snapshot.id, 'ex': e})
        for tgt_lun in tgt_luns:
            snapshots_model_update.append({
                'id': tgt_lun.os_id,
                'status': model_update['status']
            })
        return model_update, snapshots_model_update

    def _delete_lun(self, lun_ids_str):
        self._client.send('DELETE', '/volumes',
                          params={'ids': lun_ids_str})

    def delete_lun_by_id(self, lun_id):
        self._client.send('DELETE', '/volumes/%s' % lun_id)

    def _get_version(self):
        return self._client.fetchone('GET', '/systems')

    @proxy.logger
    def _create_lun(self, volData):
        return self._client.fetchid('POST', '/volumes', volData)

    def _get_pool(self, pool_id):
        return self._client.fetchone('GET', '/pools/%s' % pool_id,
                                     fields=['id', 'name', 'node', 'stgtype',
                                             'cap', 'capavail'])

    def start_flashcopy(self, vol_pairs, freeze=False):
        options = [
            "permit_space_efficient_target",
            "fail_space_efficient_target_out_of_space"
        ]
        if freeze:
            options.append("freeze_consistency")
        self._client.send('POST', '/cs/flashcopies', {
            "volume_pairs": vol_pairs,
            "options": options
        })

    def get_pprc_paths(self, specific_lss=None):
        if specific_lss:
            lss_range = {
                'source_lss_id_from': specific_lss,
                'source_lss_id_to': specific_lss
            }
        else:
            # get all of PPRC paths between source DS8K and target DS8K.
            lss_range = {
                'source_lss_id_from': '00',
                'source_lss_id_to': 'FF'
            }

        return self._client.fetchall('GET', '/cs/pprcs/paths',
                                     params=lss_range)

    def get_flashcopy(self, vol_id):
        return self._client.fetchall('GET', '/volumes/%s/flashcopy' % vol_id)

    def delete_flashcopy(self, src_lun_id, tgt_lun_id):
        # no exception if failed
        self._client.statusok(
            'DELETE', '/cs/flashcopies/%s:%s' % (src_lun_id, tgt_lun_id))

    def _get_host_ports(self, host_wwpn_set):
        return self._client.fetchall(
            'GET', '/host_ports',
            params={
                'wwpns': ",".join(host_wwpn_set),
                'state': 'logged in,logged out'
            },
            fields=['host_id', 'wwpn'])

    def _create_host(self, host):
        return self._client.fetchone(
            'POST', '/hosts', {'name': host.name, 'hosttype': host.type})

    def _assign_host_port(self, host_id, ports):
        self._client.send('POST', '/host_ports/assign', {
            'host_id': host_id, 'host_port_wwpns': ports})

    def _map_volume_to_host(self, host_id, vol_id):
        return self._client.fetchid(
            'POST', '/hosts%5Bid=' + host_id + '%5D/mappings',
            {'volumes': [vol_id]})

    def _get_mappings(self, host_id):
        return self._client.fetchall(
            'GET', '/hosts%5Bid=' + host_id + '%5D/mappings')

    def _delete_mappings(self, host_id, lun_id):
        self._client.send(
            'DELETE', '/hosts%5Bid=' + host_id + '%5D/mappings/' + lun_id)

    def _delete_host_ports(self, port):
        self._client.send('DELETE', '/host_ports/%s' % port)

    def _delete_host(self, host_id):
        # delete the host will delete all of the ports belong to it
        self._client.send('DELETE', '/hosts%5Bid=' + host_id + '%5D')

    def _get_ioports(self):
        return self._client.fetchall('GET', '/ioports', fields=['wwpn'])

    def unfreeze_lss(self, lss_ids):
        self._client.send(
            'POST', '/cs/flashcopies/unfreeze', {"lss_ids": lss_ids})

    def get_all_lss(self, fields=None):
        fields = (fields if fields else
                  ['id', 'type', 'group', 'configvols'])
        return self._client.fetchall('GET', '/lss', fields=fields)

    def lun_exists(self, lun_id):
        return self._client.statusok('GET', '/volumes/%s' % lun_id)

    def get_lun_pool(self, lun_id):
        return self._client.fetchone(
            'GET', '/volumes/%s' % lun_id, fields=['pool'])['pool']

    def change_lun(self, lun_id, param):
        self._client.send('PUT', '/volumes/%s' % lun_id, param)

    def get_physical_links(self, target_id):
        return self._client.fetchall(
            'GET', '/cs/pprcs/physical_links',
            params={
                'target_system_wwnn': target_id,
                'source_lss_id': 00,
                'target_lss_id': 00
            })

    def get_systems(self):
        return self._client.fetchone(
            'GET', '/systems', fields=['id', 'wwnn', 'release'])

    def get_lun_number_in_lss(self, lss_id):
        return int(self._client.fetchone(
            'GET', '/lss/%s' % lss_id,
            fields=['configvols'])['configvols'])

    def create_pprc_path(self, pathData):
        self._client.send('POST', '/cs/pprcs/paths', pathData)

    def get_pprc_path(self, path_id):
        return self._client.fetchone(
            'GET', '/cs/pprcs/paths/%s' % path_id,
            fields=['port_pairs'])

    def delete_pprc_path(self, path_id):
        self._client.send('DELETE', '/cs/pprcs/paths/%s' % path_id)

    def create_pprc_pair(self, pair_data):
        self._client.send('POST', '/cs/pprcs', pair_data)

    def delete_pprc_pair_by_pair_id(self, pids):
        self._client.statusok('DELETE', '/cs/pprcs', params=pids)

    def do_failback(self, pair_data):
        self._client.send('POST', '/cs/pprcs/resume', pair_data)

    def get_pprc_pairs(self, min_vol_id, max_vol_id):
        return self._client.fetchall(
            'GET', '/cs/pprcs',
            params={
                'volume_id_from': min_vol_id,
                'volume_id_to': max_vol_id
            })

    def delete_pprc_pair(self, vol_id):
        # check pprc pairs exist or not.
        if not self.get_pprc_pairs(vol_id, vol_id):
            return None
        # don't use pprc pair ID to delete it, because it may have
        # communication issues.
        pair_data = {
            'volume_full_ids': [{
                'volume_id': vol_id,
                'system_id': self.backend['storage_unit']
            }],
            'options': ['unconditional', 'issue_source']
        }
        self._client.send('POST', '/cs/pprcs/delete', pair_data)

    def pause_pprc_pairs(self, pprc_pair_ids):
        pair_data = {'pprc_ids': pprc_pair_ids}
        self._client.send('POST', '/cs/pprcs/pause', pair_data)

    def resume_pprc_pairs(self, pprc_pair_ids):
        pair_data = {
            'pprc_ids': pprc_pair_ids,
            'type': 'metro_mirror',
            'options': ['permit_space_efficient_target',
                        'initial_copy_out_of_sync']
        }
        self._client.send('POST', '/cs/pprcs/resume', pair_data)


class DS8KReplicationSourceHelper(DS8KCommonHelper):
    """Manage source storage for replication."""

    @proxy.logger
    def find_pool_and_lss(self, excluded_lss=None):
        for pool_id, pool in self._storage_pools.items():
            lss = self._find_lss_for_type_replication(pool['node'],
                                                      excluded_lss)
            if lss:
                return pool_id, lss
        raise restclient.LssIDExhaustError(
            message=_("All LSS/LCU IDs for configured pools are exhausted."))

    @proxy.logger
    def _find_lss_for_type_replication(self, node, excluded_lss):
        # prefer to choose non-existing one first.
        existing_lss = self.get_all_lss()
        LOG.info("existing LSS IDs are %s",
                 ','.join([lss['id'] for lss in existing_lss]))
        existing_lss_cg, nonexistent_lss_cg = (
            self._classify_lss_for_cg(existing_lss))
        lss_id = self._find_from_nonexistent_lss(node, existing_lss,
                                                 nonexistent_lss_cg)
        if not lss_id:
            if excluded_lss:
                existing_lss = [lss for lss in existing_lss
                                if lss['id'] not in excluded_lss]
            candidates = [lss for lss in existing_lss
                          if lss['id'] not in existing_lss_cg]
            lss_id = self._find_from_existing_lss(node, candidates)
        return lss_id


class DS8KReplicationTargetHelper(DS8KReplicationSourceHelper):
    """Manage target storage for replication."""

    OPTIONAL_PARAMS = ['ds8k_host_type', 'port_pairs', 'lss_range_for_cg']

    def setup(self):
        self._create_client()
        self._get_storage_information()
        self._get_replication_information()
        self._check_host_type()
        self.backend['lss_ids_for_cg'] = self._get_lss_ids_for_cg()
        self.backend['pools_str'] = self._get_value(
            'san_clustername').replace('_', ',')
        self._storage_pools = self.get_pools()
        self.verify_pools(self._storage_pools)
        self._verify_rest_version()

    def _get_replication_information(self):
        port_pairs = []
        pairs = self._get_value('port_pairs')
        if pairs:
            for pair in pairs.replace(' ', '').upper().split(';'):
                pair = pair.split('-')
                port_pair = {
                    'source_port_id': pair[0],
                    'target_port_id': pair[1]
                }
                port_pairs.append(port_pair)
        self.backend['port_pairs'] = port_pairs
        self.backend['id'] = self._get_value('backend_id')

    def create_lun(self, lun):
        volData = {
            'cap': self._gb2b(lun.size),
            'captype': 'bytes',
            'stgtype': 'fb',
            'tp': 'ese' if lun.type_thin else 'none'
        }
        lun.data_type = lun.data_type if lun.data_type else 'FB 512'
        if lun.type_os400:
            volData['os400'] = lun.type_os400

        volData['name'] = lun.replica_ds_name
        volData['pool'], volData['lss'] = lun.pool_lss_pair['target']
        volID = self._create_lun(volData)
        lun.replication_driver_data.update(
            {self.backend['id']: {'vol_hex_id': volID}})
        return lun

    def delete_pprc_pair(self, vol_id):
        if not self.get_pprc_pairs(vol_id, vol_id):
            return None
        pair_data = {
            'volume_full_ids': [{
                'volume_id': vol_id,
                'system_id': self.backend['storage_unit']
            }],
            'options': ['unconditional', 'issue_target']
        }
        self._client.send('POST', '/cs/pprcs/delete', pair_data)


class DS8KECKDHelper(DS8KCommonHelper):
    """Manage ECKD volume."""

    OPTIONAL_PARAMS = ['ds8k_host_type', 'port_pairs', 'ds8k_ssid_prefix',
                       'lss_range_for_cg']
    # if use new REST API, please update the version below
    VALID_REST_VERSION_87_51_MIN = '87.51.63.0'
    VALID_REST_VERSION_88_20_MIN = '88.20.112.0'
    MIN_VALID_STORAGE_VERSION = '8.1'
    INVALID_STORAGE_VERSION = '8.0.1'

    @staticmethod
    def _gb2cyl(gb):
        # now only support 3390, no 3380 or 3390-A
        cyl = int(math.ceil(gb * 1263.28))
        if cyl > 65520:
            raise exception.VolumeDriverException(
                message=(_("For 3390 volume, capacity can be in the range "
                           "1-65520(849KiB to 55.68GiB) cylinders, now it "
                           "is %(gb)d GiB, equals to %(cyl)d cylinders.")
                         % {'gb': gb, 'cyl': cyl}))
        return cyl

    @staticmethod
    def _cyl2b(cyl):
        return cyl * 849960

    def _get_cula(self, lcu):
        return self.backend['device_mapping'][lcu]

    def disable_thin_provision(self):
        self._disable_thin_provision = True

    def setup(self):
        self._create_client()
        self._get_storage_information()
        self._check_host_type()
        self.backend['lss_ids_for_cg'] = self._get_lss_ids_for_cg()
        self.backend['pools_str'] = self._get_value('san_clustername')
        self._storage_pools = self.get_pools()
        self.verify_pools(self._storage_pools)
        ssid_prefix = self._get_value('ds8k_ssid_prefix')
        self.backend['ssid_prefix'] = ssid_prefix if ssid_prefix else 'FF'
        self.backend['device_mapping'] = self._get_device_mapping()
        self._verify_rest_version()

    def _verify_rest_version(self):
        if self.backend['storage_version'] == self.INVALID_STORAGE_VERSION:
            raise exception.VolumeDriverException(
                message=(_("%s does not support bulk deletion of volumes, "
                           "if you want to use this version of driver, "
                           "please upgrade the CCL.")
                         % self.INVALID_STORAGE_VERSION))
        # DS8K supports ECKD ESE volume from 8.1
        if (dist_version.LooseVersion(self.backend['storage_version']) <
           dist_version.LooseVersion(self.MIN_VALID_STORAGE_VERSION)):
            self._disable_thin_provision = True

        rest_ver = self.backend['rest_version'][0:2]
        if (('87' == rest_ver and
           dist_version.LooseVersion(self.backend['rest_version']) <
           dist_version.LooseVersion(self.VALID_REST_VERSION_87_51_MIN)) or
           ('88' == rest_ver and
           dist_version.LooseVersion(self.backend['rest_version']) <
           dist_version.LooseVersion(self.VALID_REST_VERSION_88_20_MIN))):
            raise exception.VolumeDriverException(
                message=(_("REST version %(invalid)s is lower than "
                           "%(valid)s, please upgrade it in DS8K.")
                         % {'invalid': self.backend['rest_version'],
                            'valid': (self.VALID_REST_VERSION_87_51_MIN
                                      if '87' == rest_ver else
                                      self.VALID_REST_VERSION_88_20_MIN)}))

    @proxy.logger
    def _get_device_mapping(self):
        map_str = self._get_value('ds8k_devadd_unitadd_mapping')
        mappings = map_str.replace(' ', '').upper().split(';')
        pairs = [m.split('-') for m in mappings]
        self.verify_lss_ids(','.join([p[1] for p in pairs]))
        return {p[1]: int(p[0], 16) for p in pairs}

    @proxy.logger
    def verify_lss_ids(self, specified_lcu_ids):
        if not specified_lcu_ids:
            return None
        lcu_ids = specified_lcu_ids.upper().replace(' ', '').split(',')
        # verify the LCU ID.
        for lcu in lcu_ids:
            if int(lcu, 16) > 255:
                raise exception.InvalidParameterValue(
                    err=_('LCU %s should be within 00-FF.') % lcu)

        # verify address group
        self._existing_lss = self.get_all_lss()
        fb_addrgrps = set(int(lss['id'], 16) // 16 for lss in
                          self._existing_lss if lss['type'] == 'fb')
        ckd_addrgrps = set((int(lcu, 16) // 16) for lcu in lcu_ids)
        intersection = ckd_addrgrps & fb_addrgrps
        if intersection:
            raise exception.VolumeDriverException(
                message=_('LCUs in the address group %s are reserved '
                          'for FB volumes') % list(intersection))

        # create LCU that doesn't exist
        nonexistent_lcu = set(lcu_ids) - set(
            lss['id'] for lss in self._existing_lss if lss['type'] == 'ckd')
        if nonexistent_lcu:
            LOG.info('LCUs %s do not exist in DS8K, they will be '
                     'created.', ','.join(nonexistent_lcu))
            for lcu in nonexistent_lcu:
                try:
                    self._create_lcu(self.backend['ssid_prefix'], lcu)
                except restclient.APIException as e:
                    raise exception.VolumeDriverException(
                        message=(_('Can not create lcu %(lcu)s, '
                                   'Exception = %(e)s.')
                                 % {'lcu': lcu, 'e': six.text_type(e)}))
        return lcu_ids

    def _format_pools(self, pools):
        return ((p['id'], {
            'name': p['name'],
            'node': int(p['node']),
            'stgtype': p['stgtype'],
            'cap': self._cyl2b(int(p['cap'])),
            'capavail': self._cyl2b(int(p['capavail']))
        }) for p in pools)

    @proxy.logger
    def find_pool_and_lss(self, excluded_lss=None):
        return self.find_biggest_pool_and_lss(excluded_lss)

    @proxy.logger
    def _find_lss(self, node, excluded_lcu, specified_lcu_ids=None):
        # all LCUs have existed, unlike LSS.
        if specified_lcu_ids:
            for lcu_id in specified_lcu_ids:
                if lcu_id not in self.backend['device_mapping'].keys():
                    raise exception.InvalidParameterValue(
                        err=_("LCU %s is not in parameter "
                              "ds8k_devadd_unitadd_mapping, "
                              "Please specify LCU in it, otherwise "
                              "driver can not attach volume.") % lcu_id)
            all_lss = self._existing_lss
        else:
            all_lss = self.get_all_lss()
        existing_lcu = [lcu for lcu in all_lss if
                        lcu['type'] == 'ckd' and
                        lcu['id'] in self.backend['device_mapping'].keys() and
                        lcu['group'] == six.text_type(node)]
        LOG.info("All appropriate LCUs are %s.",
                 ','.join([lcu['id'] for lcu in existing_lcu]))

        # exclude full LCUs.
        if excluded_lcu:
            existing_lcu = [lcu for lcu in existing_lcu if
                            lcu['id'] not in excluded_lcu]
            if not existing_lcu:
                LOG.info("All appropriate LCUs are full.")
                return None

        ignore_pprc = False
        if specified_lcu_ids:
            # user specify LCUs in extra-specs.
            existing_lcu = [lcu for lcu in existing_lcu
                            if lcu['id'] in specified_lcu_ids]
            ignore_pprc = True

        # exclude LCUs reserved for CG.
        existing_lcu = [lcu for lcu in existing_lcu if lcu['id']
                        not in self.backend['lss_ids_for_cg']]
        if not existing_lcu:
            LOG.info("All appropriate LCUs have been reserved for "
                     "for consistency group.")
            return None

        if not ignore_pprc:
            # prefer to use LCU that is not in PPRC path first.
            lcu_pprc = self.get_lss_in_pprc_paths() & set(
                self.backend['device_mapping'].keys())
            if lcu_pprc:
                lcu_non_pprc = [
                    lcu for lcu in existing_lcu if lcu['id'] not in lcu_pprc]
                if lcu_non_pprc:
                    existing_lcu = lcu_non_pprc

        # return LCU which has max number of empty slots.
        emptiest_lcu = sorted(
            existing_lcu, key=lambda i: int(i['configvols']))[0]
        if int(emptiest_lcu['configvols']) == LSS_VOL_SLOTS:
            return None
        else:
            return emptiest_lcu['id']

    def _create_lcu(self, ssid_prefix, lcu):
        self._client.send('POST', '/lss', {
            'id': lcu,
            'type': 'ckd',
            'sub_system_identifier': ssid_prefix + lcu
        })

    def create_lun(self, lun):
        volData = {
            'cap': self._gb2cyl(lun.size),
            'captype': 'cyl',
            'stgtype': 'ckd',
            'tp': 'ese' if lun.type_thin else 'none'
        }
        lun.data_type = '3390'
        volData['name'] = lun.ds_name
        volData['pool'], volData['lss'] = lun.pool_lss_pair['source']
        lun.ds_id = self._create_lun(volData)
        return lun

    def initialize_connection(self, vol_id, connector, **kwargs):
        return {
            'driver_volume_type': 'fibre_channel_eckd',
            'data': {
                'target_discovered': True,
                'cula': self._get_cula(vol_id[0:2]),
                'unit_address': int(vol_id[2:4], 16),
                'discard': False
            }
        }

    def terminate_connection(self, vol_id, connector, force, **kwargs):
        return None


class DS8KReplicationTargetECKDHelper(DS8KECKDHelper,
                                      DS8KReplicationTargetHelper):
    """Manage ECKD volume in replication target."""

    def setup(self):
        self._create_client()
        self._get_storage_information()
        self._get_replication_information()
        self._check_host_type()
        self.backend['lss_ids_for_cg'] = self._get_lss_ids_for_cg()
        self.backend['pools_str'] = self._get_value(
            'san_clustername').replace('_', ',')
        self._storage_pools = self.get_pools()
        self.verify_pools(self._storage_pools)
        ssid_prefix = self._get_value('ds8k_ssid_prefix')
        self.backend['ssid_prefix'] = ssid_prefix if ssid_prefix else 'FF'
        self.backend['device_mapping'] = self._get_device_mapping()
        self._verify_rest_version()

    def create_lun(self, lun):
        volData = {
            'cap': self._gb2cyl(lun.size),
            'captype': 'cyl',
            'stgtype': 'ckd',
            'tp': 'ese' if lun.type_thin else 'none'
        }
        lun.data_type = '3390'

        volData['name'] = lun.replica_ds_name
        volData['pool'], volData['lss'] = lun.pool_lss_pair['target']
        volID = self._create_lun(volData)
        lun.replication_driver_data.update(
            {self.backend['id']: {'vol_hex_id': volID}})
        return lun



# Copyright 2012 Pedro Navarro Perez
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Volume driver for Windows Server 2012

This driver requires ISCSI target role installed

"""

import contextlib
import os

from os_win import utilsfactory
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import fileutils
from oslo_utils import units
from oslo_utils import uuidutils

from cinder import exception
from cinder.image import image_utils
from cinder import interface
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)

windows_opts = [
    cfg.StrOpt('windows_iscsi_lun_path',
               default=r'C:\iSCSIVirtualDisks',
               help='Path to store VHD backed volumes'),
]

CONF = cfg.CONF
CONF.register_opts(windows_opts, group=configuration.SHARED_CONF_GROUP)


@interface.volumedriver
class WindowsISCSIDriver(driver.ISCSIDriver):
    """Executes volume driver commands on Windows Storage server."""

    VERSION = '1.0.0'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Microsoft_iSCSI_CI"

    def __init__(self, *args, **kwargs):
        super(WindowsISCSIDriver, self).__init__(*args, **kwargs)
        self.configuration = kwargs.get('configuration', None)
        if self.configuration:
            self.configuration.append_config_values(windows_opts)

        self._vhdutils = utilsfactory.get_vhdutils()
        self._tgt_utils = utilsfactory.get_iscsi_target_utils()
        self._hostutils = utilsfactory.get_hostutils()

    @staticmethod
    def get_driver_options():
        return windows_opts

    def do_setup(self, context):
        """Setup the Windows Volume driver.

        Called one time by the manager after the driver is loaded.
        Validate the flags we care about
        """
        fileutils.ensure_tree(self.configuration.windows_iscsi_lun_path)
        fileutils.ensure_tree(CONF.image_conversion_dir)

    def check_for_setup_error(self):
        """Check that the driver is working and can communicate."""
        self._get_portals()

    def _get_portals(self):
        available_portals = set(self._tgt_utils.get_portal_locations(
            available_only=True,
            fail_if_none_found=True))
        LOG.debug("Available iSCSI portals: %s", available_portals)

        iscsi_port = self.configuration.target_port
        iscsi_ips = ([self.configuration.target_ip_address] +
                     self.configuration.iscsi_secondary_ip_addresses)
        requested_portals = {':'.join([iscsi_ip, str(iscsi_port)])
                             for iscsi_ip in iscsi_ips}

        unavailable_portals = requested_portals - available_portals
        if unavailable_portals:
            LOG.warning("The following iSCSI portals were requested but "
                        "are not available: %s.", unavailable_portals)

        selected_portals = requested_portals & available_portals
        if not selected_portals:
            err_msg = "None of the configured iSCSI portals are available."
            raise exception.VolumeDriverException(err_msg)

        return list(selected_portals)

    def _get_host_information(self, volume, multipath=False):
        """Getting the portal and port information."""
        target_name = self._get_target_name(volume)

        available_portals = self._get_portals()
        properties = self._tgt_utils.get_target_information(target_name)

        # Note(lpetrut): the WT_Host CHAPSecret field cannot be accessed
        # for security reasons.
        auth = volume.provider_auth
        if auth:
            (auth_method, auth_username, auth_secret) = auth.split()
            properties['auth_method'] = auth_method
            properties['auth_username'] = auth_username
            properties['auth_password'] = auth_secret

        properties['target_portal'] = available_portals[0]
        properties['target_discovered'] = False
        properties['target_lun'] = 0
        properties['volume_id'] = volume.id

        if multipath:
            properties['target_portals'] = available_portals
            properties['target_iqns'] = [properties['target_iqn']
                                         for portal in available_portals]
            properties['target_luns'] = [properties['target_lun']
                                         for portal in available_portals]

        return properties

    def initialize_connection(self, volume, connector):
        """Driver entry point to attach a volume to an instance."""
        initiator_name = connector['initiator']
        target_name = volume.provider_location

        self._tgt_utils.associate_initiator_with_iscsi_target(initiator_name,
                                                              target_name)

        properties = self._get_host_information(volume,
                                                connector.get('multipath'))

        return {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }

    def terminate_connection(self, volume, connector, **kwargs):
        """Driver entry point to unattach a volume from an instance.

        Unmask the LUN on the storage system so the given initiator can no
        longer access it.
        """
        initiator_name = connector['initiator']
        target_name = volume.provider_location
        self._tgt_utils.deassociate_initiator(initiator_name, target_name)

    def create_volume(self, volume):
        """Driver entry point for creating a new volume."""
        vhd_path = self.local_path(volume)
        vol_name = volume.name
        vol_size_mb = volume.size * 1024

        self._tgt_utils.create_wt_disk(vhd_path, vol_name,
                                       size_mb=vol_size_mb)

    def local_path(self, volume, disk_format=None):
        base_vhd_folder = self.configuration.windows_iscsi_lun_path
        if not disk_format:
            disk_format = self._tgt_utils.get_supported_disk_format()

        disk_fname = "%s.%s" % (volume.name, disk_format)
        return os.path.join(base_vhd_folder, disk_fname)

    def delete_volume(self, volume):
        """Driver entry point for destroying existing volumes."""
        vol_name = volume.name
        vhd_path = self.local_path(volume)

        self._tgt_utils.remove_wt_disk(vol_name)
        fileutils.delete_if_exists(vhd_path)

    def create_snapshot(self, snapshot):
        """Driver entry point for creating a snapshot."""
        # Getting WT_Snapshot class
        vol_name = snapshot.volume_name
        snapshot_name = snapshot.name

        self._tgt_utils.create_snapshot(vol_name, snapshot_name)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Driver entry point for exporting snapshots as volumes."""
        snapshot_name = snapshot.name
        vol_name = volume.name
        vhd_path = self.local_path(volume)

        self._tgt_utils.export_snapshot(snapshot_name, vhd_path)
        self._tgt_utils.import_wt_disk(vhd_path, vol_name)

    def delete_snapshot(self, snapshot):
        """Driver entry point for deleting a snapshot."""
        snapshot_name = snapshot.name
        self._tgt_utils.delete_snapshot(snapshot_name)

    def ensure_export(self, context, volume):
        # iSCSI targets exported by WinTarget persist after host reboot.
        pass

    def _get_target_name(self, volume):
        return "%s%s" % (self.configuration.target_prefix,
                         volume.name)

    def create_export(self, context, volume, connector):
        """Driver entry point to get the export info for a new volume."""
        target_name = self._get_target_name(volume)
        updates = {}

        if not self._tgt_utils.iscsi_target_exists(target_name):
            self._tgt_utils.create_iscsi_target(target_name)
            updates['provider_location'] = target_name

            if self.configuration.use_chap_auth:
                chap_username = (self.configuration.chap_username or
                                 volume_utils.generate_username())
                chap_password = (self.configuration.chap_password or
                                 volume_utils.generate_password())

                self._tgt_utils.set_chap_credentials(target_name,
                                                     chap_username,
                                                     chap_password)

                updates['provider_auth'] = ' '.join(('CHAP',
                                                     chap_username,
                                                     chap_password))

        # This operation is idempotent
        self._tgt_utils.add_disk_to_target(volume.name, target_name)

        return updates

    def remove_export(self, context, volume):
        """Driver entry point to remove an export for a volume."""
        target_name = self._get_target_name(volume)
        self._tgt_utils.delete_iscsi_target(target_name)

    def copy_image_to_volume(self, context, volume, image_service, image_id):
        """Fetch the image from image_service and create a volume using it."""
        # Convert to VHD and file back to VHD
        vhd_type = self._tgt_utils.get_supported_vhd_type()
        with image_utils.temporary_file(suffix='.vhd') as tmp:
            volume_path = self.local_path(volume)
            image_utils.fetch_to_vhd(context, image_service, image_id, tmp,
                                     self.configuration.volume_dd_blocksize)
            # The vhd must be disabled and deleted before being replaced with
            # the desired image.
            self._tgt_utils.change_wt_disk_status(volume.name,
                                                  enabled=False)
            os.unlink(volume_path)
            self._vhdutils.convert_vhd(tmp, volume_path,
                                       vhd_type)
            self._vhdutils.resize_vhd(volume_path,
                                      volume.size << 30,
                                      is_file_max_size=False)
            self._tgt_utils.change_wt_disk_status(volume.name,
                                                  enabled=True)

    @contextlib.contextmanager
    def _temporary_snapshot(self, volume_name):
        try:
            snap_uuid = uuidutils.generate_uuid()
            snapshot_name = '%s-tmp-snapshot-%s' % (volume_name, snap_uuid)
            self._tgt_utils.create_snapshot(volume_name, snapshot_name)
            yield snapshot_name
        finally:
            self._tgt_utils.delete_snapshot(snapshot_name)

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        """Copy the volume to the specified image."""
        disk_format = self._tgt_utils.get_supported_disk_format()
        temp_vhd_path = os.path.join(CONF.image_conversion_dir,
                                     str(image_meta['id']) + '.' + disk_format)

        try:
            with self._temporary_snapshot(volume.name) as tmp_snap_name:
                # qemu-img cannot access VSS snapshots, for which reason it
                # must be exported first.
                self._tgt_utils.export_snapshot(tmp_snap_name, temp_vhd_path)
                volume_utils.upload_volume(
                    context, image_service, image_meta, temp_vhd_path, volume,
                    'vhd')
        finally:
            fileutils.delete_if_exists(temp_vhd_path)

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""
        src_vol_name = src_vref.name
        vol_name = volume.name
        vol_size = volume.size

        new_vhd_path = self.local_path(volume)

        with self._temporary_snapshot(src_vol_name) as tmp_snap_name:
            self._tgt_utils.export_snapshot(tmp_snap_name, new_vhd_path)
            self._vhdutils.resize_vhd(new_vhd_path, vol_size << 30,
                                      is_file_max_size=False)

            self._tgt_utils.import_wt_disk(new_vhd_path, vol_name)

    def _get_capacity_info(self):
        drive = os.path.splitdrive(
            self.configuration.windows_iscsi_lun_path)[0]
        (size, free_space) = self._hostutils.get_volume_info(drive)

        total_gb = size / units.Gi
        free_gb = free_space / units.Gi
        return (total_gb, free_gb)

    def _update_volume_stats(self):
        """Retrieve stats info for Windows device."""
        LOG.debug("Updating volume stats")
        total_gb, free_gb = self._get_capacity_info()

        data = {}
        backend_name = self.configuration.safe_get('volume_backend_name')
        data["volume_backend_name"] = backend_name or self.__class__.__name__
        data["vendor_name"] = 'Microsoft'
        data["driver_version"] = self.VERSION
        data["storage_protocol"] = 'iSCSI'
        data['total_capacity_gb'] = total_gb
        data['free_capacity_gb'] = free_gb
        data['reserved_percentage'] = self.configuration.reserved_percentage
        data['QoS_support'] = False

        self._stats = data

    def extend_volume(self, volume, new_size):
        """Extend an Existing Volume."""
        old_size = volume.size
        LOG.debug("Extend volume from %(old_size)s GB to %(new_size)s GB.",
                  {'old_size': old_size, 'new_size': new_size})
        additional_size_mb = (new_size - old_size) * 1024

        self._tgt_utils.extend_wt_disk(volume.name, additional_size_mb)

    def backup_use_temp_snapshot(self):
        return False



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Policy Engine For Cinder"""

import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import opts as policy_opts
from oslo_policy import policy
from oslo_utils import excutils

from cinder import exception
from cinder import policies

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
DEFAULT_POLICY_FILENAME = 'policy.yaml'
policy_opts.set_defaults(cfg.CONF, DEFAULT_POLICY_FILENAME)

_ENFORCER = None


def reset():
    global _ENFORCER
    if _ENFORCER:
        _ENFORCER.clear()
        _ENFORCER = None


def init(use_conf=True,
         suppress_deprecation_warnings=False):
    """Init an Enforcer class.

    :param use_conf: Whether to load rules from config file.
    """

    global _ENFORCER
    if not _ENFORCER:
        _ENFORCER = policy.Enforcer(
            CONF,
            use_conf=use_conf,
            fallback_to_json_file=False)
        _ENFORCER.suppress_deprecation_warnings = suppress_deprecation_warnings
        register_rules(_ENFORCER)
        _ENFORCER.load_rules()


def enforce(context, action, target):
    """Verifies that the action is valid on the target in this context.

    :param context: cinder context
    :param action: string representing the action to be checked
                   this should be colon separated for clarity.
                   i.e. ``compute:create_instance``,
                   ``compute:attach_volume``,
                   ``volume:attach_volume``

    :param target: dictionary representing the object of the action for object
                   creation this should be a dictionary representing the
                   location of the object e.g.
                   ``{'project_id': context.project_id}``

    :raises PolicyNotAuthorized: if verification fails.
    """
    init()

    try:
        return _ENFORCER.enforce(action,
                                 target,
                                 context,
                                 do_raise=True,
                                 exc=exception.PolicyNotAuthorized,
                                 action=action)
    except policy.InvalidScope:
        raise exception.PolicyNotAuthorized(action=action)


def set_rules(rules, overwrite=True, use_conf=False):
    """Set rules based on the provided dict of rules.

    :param rules: New rules to use. It should be an instance of dict.
    :param overwrite: Whether to overwrite current rules or update them
                      with the new rules.
    :param use_conf: Whether to reload rules from config file.
    """

    init(use_conf=False)
    _ENFORCER.set_rules(rules, overwrite, use_conf)


def get_rules():
    if _ENFORCER:
        return _ENFORCER.rules


def register_rules(enforcer):
    enforcer.register_defaults(policies.list_rules())


def get_enforcer():
    # This method is for use by oslopolicy CLI scripts. Those scripts need the
    # 'output-file' and 'namespace' options, but having those in sys.argv means
    # loading the Cinder config options will fail as those are not expected to
    # be present. So we pass in an arg list with those stripped out.
    conf_args = []
    # Start at 1 because cfg.CONF expects the equivalent of sys.argv[1:]
    i = 1
    while i < len(sys.argv):
        if sys.argv[i].strip('-') in ['namespace', 'output-file']:
            i += 2
            continue
        conf_args.append(sys.argv[i])
        i += 1

    cfg.CONF(conf_args, project='cinder')
    init()
    return _ENFORCER


def authorize(context, action, target, do_raise=True, exc=None):
    """Verifies that the action is valid on the target in this context.

    :param context: cinder context
    :param action: string representing the action to be checked
                   this should be colon separated for clarity.
                   i.e. ``compute:create_instance``,
                   ``compute:attach_volume``,
                   ``volume:attach_volume``
    :param target: dictionary representing the object of the action for object
                   creation this should be a dictionary representing the
                   location of the object e.g.
                   ``{'project_id': context.project_id}``
    :param do_raise: if True (the default), raises PolicyNotAuthorized;
                     if False, returns False
    :param exc: Class of the exception to raise if the check fails.
                Any remaining arguments passed to :meth:`authorize` (both
                positional and keyword arguments) will be passed to
                the exception class. If not specified,
                :class:`PolicyNotAuthorized` will be used.

    :raises cinder.exception.PolicyNotAuthorized: if verification fails
           and do_raise is True. Or if 'exc' is specified it will raise an
           exception of that type.

    :return: returns a non-False value (not necessarily "True") if
             authorized, and the exact value False if not authorized and
             do_raise is False.
    """
    init()
    if not exc:
        exc = exception.PolicyNotAuthorized
    try:
        result = _ENFORCER.authorize(action, target, context,
                                     do_raise=do_raise, exc=exc, action=action)
    except policy.PolicyNotRegistered:
        with excutils.save_and_reraise_exception():
            LOG.exception('Policy not registered')
    except Exception:
        with excutils.save_and_reraise_exception():
            LOG.debug('Policy check for %(action)s failed with context '
                      '%(credentials)s',
                      {'action': action,
                       'credentials': context.to_policy_values()})
    return result


def check_is_admin(context):
    """Whether or not user is admin according to policy setting."""
    init()
    # the target is user-self
    credentials = context.to_policy_values()
    target = credentials
    return _ENFORCER.authorize('context_is_admin', target, credentials)



# Copyright 2020 Datera
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import uuid

from eventlet.green import threading
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
import six

from cinder import exception
from cinder.i18n import _
from cinder import interface
import cinder.volume.drivers.datera.datera_api21 as api21
import cinder.volume.drivers.datera.datera_api22 as api22
import cinder.volume.drivers.datera.datera_common as datc
from cinder.volume.drivers.san import san
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)

dfs_sdk = importutils.try_import('dfs_sdk')

d_opts = [
    cfg.StrOpt('datera_api_port',
               default='7717',
               deprecated_for_removal=True,
               help='Datera API port.'),
    cfg.StrOpt('datera_api_version',
               default='2.2',
               deprecated_for_removal=True,
               help='Datera API version.'),
    cfg.StrOpt('datera_ldap_server',
               default=None,
               help='LDAP authentication server'),
    cfg.IntOpt('datera_503_timeout',
               default='120',
               help='Timeout for HTTP 503 retry messages'),
    cfg.IntOpt('datera_503_interval',
               default='5',
               help='Interval between 503 retries'),
    cfg.BoolOpt('datera_debug',
                default=False,
                help="True to set function arg and return logging"),
    cfg.BoolOpt('datera_debug_replica_count_override',
                default=False,
                help="ONLY FOR DEBUG/TESTING PURPOSES\n"
                     "True to set replica_count to 1"),
    cfg.StrOpt('datera_tenant_id',
               default=None,
               help="If set to 'Map' --> OpenStack project ID will be mapped "
                    "implicitly to Datera tenant ID\n"
                    "If set to None --> Datera tenant ID will not be used "
                    "during volume provisioning\n"
                    "If set to anything else --> Datera tenant ID will be the "
                    "provided value"),
    cfg.BoolOpt('datera_enable_image_cache',
                default=False,
                help="Set to True to enable Datera backend image caching"),
    cfg.StrOpt('datera_image_cache_volume_type_id',
               default=None,
               help="Cinder volume type id to use for cached volumes"),
    cfg.BoolOpt('datera_disable_profiler',
                default=False,
                help="Set to True to disable profiling in the Datera driver"),
    cfg.BoolOpt('datera_disable_extended_metadata',
                default=False,
                help="Set to True to disable sending additional metadata to "
                     "the Datera backend"),
    cfg.BoolOpt('datera_disable_template_override',
                default=False,
                help="Set to True to disable automatic template override of "
                     "the size attribute when creating from a template"),
    cfg.DictOpt('datera_volume_type_defaults',
                default={},
                help="Settings here will be used as volume-type defaults if "
                     "the volume-type setting is not provided.  This can be "
                     "used, for example, to set a very low total_iops_max "
                     "value if none is specified in the volume-type to "
                     "prevent accidental overusage.  Options are specified "
                     "via the following format, WITHOUT ANY 'DF:' PREFIX: "
                     "'datera_volume_type_defaults="
                     "iops_per_gb:100,bandwidth_per_gb:200...etc'."),
]


CONF = cfg.CONF
CONF.import_opt('driver_use_ssl', 'cinder.volume.driver')
CONF.register_opts(d_opts)


@six.add_metaclass(volume_utils.TraceWrapperWithABCMetaclass)
@interface.volumedriver
class DateraDriver(san.SanISCSIDriver, api21.DateraApi, api22.DateraApi):
    """The OpenStack Datera iSCSI volume driver.

    .. code-block:: none

    Version history:
        * 1.0 - Initial driver
        * 1.1 - Look for lun-0 instead of lun-1.
        * 2.0 - Update For Datera API v2
        * 2.1 - Multipath, ACL and reorg
        * 2.2 - Capabilites List, Extended Volume-Type Support
                Naming convention change,
                Volume Manage/Unmanage support
        * 2.3 - Templates, Tenants, Snapshot Polling,
                2.1 Api Version Support, Restructure
        * 2.3.1 - Scalability bugfixes
        * 2.3.2 - Volume Placement, ACL multi-attach bugfix
        * 2.4.0 - Fast Retype Support
        * 2.5.0 - Glance Image Caching, retyping/QoS bugfixes
        * 2.6.0 - Api 2.2 support
        * 2.6.1 - Glance interoperability fix
        * 2.7.0 - IOPS/GB and BW/GB settings, driver level overrides
                  (API 2.1+ only)
        * 2.7.2 - Allowing DF: QoS Spec prefix, QoS type leak bugfix
        * 2.7.3 - Fixed bug in clone_image where size was not set correctly
        * 2.7.4 - Fix for create_tenant incorrect API call
                  Temporary fix for DAT-15931
        * 2.7.5 - Removed "force" parameter from /initiators v2.1 API requests
        * 2.8.0 - iops_per_gb and bandwidth_per_gb are now limited by
                  total_iops_max and total_bandwidth_max (API 2.1+ only)
                  Bugfix for cinder retype with online volume
        * 2.8.1 - Bugfix for missing default dict during retype
        * 2.8.2 - Updated most retype operations to not detach volume
        * 2.8.3 - Bugfix for not allowing fast clones for shared/community
                  volumes
        * 2.8.4 - Fixed missing API version pinning in _offline_flip
        * 2.8.5 - Membership check for fast image cloning. Metadata API pinning
        * 2.8.6 - Added LDAP support and CHAP support
        * 2.8.7 - Bugfix for missing tenancy calls in offline_flip
        * 2.9.0 - Volumes now correctly renamed during backend migration.
                  Implemented update_migrated_volume (API 2.1+ only),
                  Prevent non-raw image cloning
        * 2.9.1 - Added extended metadata attributes during volume creation
                  and attachment.  Added datera_disable_extended_metadata
                  option to disable it.
        * 2.9.2 - Made ensure_export a no-op.  Removed usage of
                  initiator-groups
        * 2018.4.5.0 - Switch to new date-based versioning scheme.  Removed v2
                       API support
        * 2018.4.17.1 - Bugfixes to IP Pools, Templates and Initiators
        * 2018.4.25.0 - Snapshot Manage.  List Manageable Snapshots support
        * 2018.4.27.0 - Major driver revamp/restructure, no functionality
                        change
        * 2018.5.1.0 - Bugfix for Map tenant auto-creation
        * 2018.5.18.0 - Bugfix for None tenant handling
        * 2018.6.7.0 - Bugfix for missing project_id during image clone
        * 2018.7.13.0 - Massive update porting to use the Datera Python-SDK
        * 2018.7.20.0 - Driver now includes display_name in created backend
                        app_instances.
        * 2018.9.17.0 - Requirements and doc changes
        * 2018.10.8.0 - Added extra_headers to Python-SDK constructor call.
                        This allows for the SDK to send the type of driver
                        performing each request along with the request. This
                        functionality existed before the Python-SDK revamp, so
                        this change adds the functionality back in.
        * 2018.10.8.1 - Adding thread_local to Python-SDK constructor call.
                        This preserves trace_id in the logs
        * 2018.10.30.0 - Adding template_override support.  Added
                         datera_disable_template_override cfgOpt to disable
                         this feature. Updated required requests version to
                         >=2.20.0 because of a security vulnerability in
                         <=2.19.X. Added support for filter_function and
                         goodness_function.
        * 2018.11.1.0 - Adding flash and hybrid capacity info to
                        get_volume_stats
        * 2018.11.8.0 - Fixing bug that broke 2.2.X support
        * 2018.11.14.0 - Bugfixes for v2.1 API support and unicode character
                         support
        * 2019.1.24.0 - Python-SDK requirements update, README updates
        * 2019.2.25.0 - Scalability fixes and utility script updates
        * 2019.6.4.1 - Added Pypi packaging installation support
        * 2019.12.10.0 - Python 3.x support, tox tests, CI ready, live
                         migration support, image cache, bugfixes.
    """

    VERSION = '2019.12.10.0'

    CI_WIKI_NAME = "datera-ci"

    HEADER_DATA = {'Datera-Driver': 'OpenStack-Cinder-{}'.format(VERSION)}

    def __init__(self, *args, **kwargs):
        super(DateraDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(d_opts)
        self.username = self.configuration.san_login
        self.password = self.configuration.san_password
        self.ldap = self.configuration.datera_ldap_server
        self.cluster_stats = {}
        self.datera_api_token = None
        self.interval = self.configuration.datera_503_interval
        self.retry_attempts = (self.configuration.datera_503_timeout /
                               self.interval)
        self.driver_prefix = str(uuid.uuid4())[:4]
        self.datera_debug = self.configuration.datera_debug

        if self.datera_debug:
            volume_utils.setup_tracing(['method'])
        self.tenant_id = self.configuration.datera_tenant_id
        if self.tenant_id is None:
            self.tenant_id = ''
        self.defaults = self.configuration.datera_volume_type_defaults
        if self.tenant_id and self.tenant_id.lower() == 'none':
            self.tenant_id = None
        self.template_override = (
            not self.configuration.datera_disable_template_override)
        self.api_check = time.time()
        self.api_cache = []
        self.api_timeout = 0
        self.do_profile = not self.configuration.datera_disable_profiler
        self.do_metadata = (
            not self.configuration.datera_disable_extended_metadata)
        self.image_cache = self.configuration.datera_enable_image_cache
        self.image_type = self.configuration.datera_image_cache_volume_type_id
        self.thread_local = threading.local()  # pylint: disable=no-member
        self.datera_version = None
        self.apiv = None
        self.api = None
        self.filterf = self.get_filter_function()
        self.goodnessf = self.get_goodness_function()

        self.use_chap_auth = self.configuration.use_chap_auth
        self.chap_username = self.configuration.chap_username
        self.chap_password = self.configuration.chap_password

        backend_name = self.configuration.safe_get(
            'volume_backend_name')
        self.backend_name = backend_name or 'Datera'
        datc.register_driver(self)

    def do_setup(self, context):
        # If we can't authenticate through the old and new method, just fail
        # now.
        if not all([self.username, self.password]):
            msg = _("san_login and/or san_password is not set for Datera "
                    "driver in the cinder.conf. Set this information and "
                    "start the cinder-volume service again.")
            LOG.error(msg)
            raise exception.InvalidInput(msg)

        # Try each valid api version starting with the latest until we find
        # one that works
        for apiv in reversed(datc.API_VERSIONS):
            try:
                api = dfs_sdk.get_api(self.configuration.san_ip,
                                      self.username,
                                      self.password,
                                      'v{}'.format(apiv),
                                      disable_log=True,
                                      extra_headers=self.HEADER_DATA,
                                      thread_local=self.thread_local,
                                      ldap_server=self.ldap)
                system = api.system.get()
                LOG.debug('Connected successfully to cluster: %s', system.name)
                self.api = api
                self.apiv = apiv
                break
            except Exception as e:
                LOG.warning(e)

    # =================

    # =================
    # = Create Volume =
    # =================

    @datc.lookup
    def create_volume(self, volume):
        """Create a logical volume."""
        pass

    # =================
    # = Extend Volume =
    # =================

    @datc.lookup
    def extend_volume(self, volume, new_size):
        pass

    # =================

    # =================
    # = Cloned Volume =
    # =================

    @datc.lookup
    def create_cloned_volume(self, volume, src_vref):
        pass

    # =================
    # = Delete Volume =
    # =================

    @datc.lookup
    def delete_volume(self, volume):
        pass

    # =================
    # = Ensure Export =
    # =================

    @datc.lookup
    def ensure_export(self, context, volume, connector=None):
        """Gets the associated account, retrieves CHAP info and updates."""

    # =========================
    # = Initialize Connection =
    # =========================

    @datc.lookup
    def initialize_connection(self, volume, connector):
        pass

    # =================
    # = Create Export =
    # =================

    @datc.lookup
    def create_export(self, context, volume, connector):
        pass

    # =================
    # = Detach Volume =
    # =================

    @datc.lookup
    def detach_volume(self, context, volume, attachment=None):
        pass

    # ===================
    # = Create Snapshot =
    # ===================

    @datc.lookup
    def create_snapshot(self, snapshot):
        pass

    # ===================
    # = Delete Snapshot =
    # ===================

    @datc.lookup
    def delete_snapshot(self, snapshot):
        pass

    # ========================
    # = Volume From Snapshot =
    # ========================

    @datc.lookup
    def create_volume_from_snapshot(self, volume, snapshot):
        pass

    # ==========
    # = Retype =
    # ==========

    @datc.lookup
    def retype(self, ctxt, volume, new_type, diff, host):
        """Convert the volume to be of the new type.

        Returns a boolean indicating whether the retype occurred.
        :param ctxt: Context
        :param volume: A dictionary describing the volume to migrate
        :param new_type: A dictionary describing the volume type to convert to
        :param diff: A dictionary with the difference between the two types
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities (Not Used).
        """
        pass

    # ==========
    # = Manage =
    # ==========

    @datc.lookup
    def manage_existing(self, volume, existing_ref):
        """Manage an existing volume on the Datera backend

        The existing_ref must be either the current name or Datera UUID of
        an app_instance on the Datera backend in a colon separated list with
        the storage instance name and volume name.  This means only
        single storage instances and single volumes are supported for
        managing by cinder.

        Eg.

        (existing_ref['source-name'] ==
             tenant:app_inst_name:storage_inst_name:vol_name)
        if using Datera 2.1 API

        or

        (existing_ref['source-name'] ==
             app_inst_name:storage_inst_name:vol_name)

        if using 2.0 API

        :param volume:       Cinder volume to manage
        :param existing_ref: Driver-specific information used to identify a
                             volume
        """
        pass

    @datc.lookup
    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Brings an existing backend storage object under Cinder management.

        existing_ref is passed straight through from the API request's
        manage_existing_ref value, and it is up to the driver how this should
        be interpreted.  It should be sufficient to identify a storage object
        that the driver should somehow associate with the newly-created cinder
        snapshot structure.

        There are two ways to do this:

        1. Rename the backend storage object so that it matches the
           snapshot['name'] which is how drivers traditionally map between a
           cinder snapshot and the associated backend storage object.

        2. Place some metadata on the snapshot, or somewhere in the backend,
           that allows other driver requests (e.g. delete) to locate the
           backend storage object when required.

        If the existing_ref doesn't make sense, or doesn't refer to an existing
        backend storage object, raise a ManageExistingInvalidReference
        exception.

        :param snapshot:     Cinder volume snapshot to manage
        :param existing_ref: Driver-specific information used to identify a
                             volume snapshot
        """
        pass

    # ===================
    # = Manage Get Size =
    # ===================

    @datc.lookup
    def manage_existing_get_size(self, volume, existing_ref):
        """Get the size of an unmanaged volume on the Datera backend

        The existing_ref must be either the current name or Datera UUID of
        an app_instance on the Datera backend in a colon separated list with
        the storage instance name and volume name.  This means only
        single storage instances and single volumes are supported for
        managing by cinder.

        Eg.

        existing_ref == app_inst_name:storage_inst_name:vol_name

        :param volume:       Cinder volume to manage
        :param existing_ref: Driver-specific information used to identify a
                             volume on the Datera backend
        """
        pass

    @datc.lookup
    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return size of snapshot to be managed by manage_existing.

        When calculating the size, round up to the next GB.

        :param snapshot:     Cinder volume snapshot to manage
        :param existing_ref: Driver-specific information used to identify a
                             volume snapshot
        :returns size:       Volume snapshot size in GiB (integer)
        """
        pass

    # =========================
    # = Get Manageable Volume =
    # =========================

    @datc.lookup
    def get_manageable_volumes(self, cinder_volumes, marker, limit, offset,
                               sort_keys, sort_dirs):
        """List volumes on the backend available for management by Cinder.

        Returns a list of dictionaries, each specifying a volume in the host,
        with the following keys:
        - reference (dictionary): The reference for a volume, which can be
          passed to "manage_existing".
        - size (int): The size of the volume according to the storage
          backend, rounded up to the nearest GB.
        - safe_to_manage (boolean): Whether or not this volume is safe to
          manage according to the storage backend. For example, is the volume
          in use or invalid for any reason.
        - reason_not_safe (string): If safe_to_manage is False, the reason why.
        - cinder_id (string): If already managed, provide the Cinder ID.
        - extra_info (string): Any extra information to return to the user

        :param cinder_volumes: A list of volumes in this host that Cinder
                               currently manages, used to determine if
                               a volume is manageable or not.
        :param marker:    The last item of the previous page; we return the
                          next results after this value (after sorting)
        :param limit:     Maximum number of items to return
        :param offset:    Number of items to skip after marker
        :param sort_keys: List of keys to sort results by (valid keys are
                          'identifier' and 'size')
        :param sort_dirs: List of directions to sort by, corresponding to
                          sort_keys (valid directions are 'asc' and 'desc')
        """
        pass

    # ============================
    # = Get Manageable Snapshots =
    # ============================

    @datc.lookup
    def get_manageable_snapshots(self, cinder_snapshots, marker, limit,
                                 offset, sort_keys, sort_dirs):
        """List snapshots on the backend available for management by Cinder.

        Returns a list of dictionaries, each specifying a snapshot in the host,
        with the following keys:
        - reference (dictionary): The reference for a snapshot, which can be
        passed to "manage_existing_snapshot".
        - size (int): The size of the snapshot according to the storage
        backend, rounded up to the nearest GB.
        - safe_to_manage (boolean): Whether or not this snapshot is safe to
        manage according to the storage backend. For example, is the snapshot
        in use or invalid for any reason.
        - reason_not_safe (string): If safe_to_manage is False, the reason why.
        - cinder_id (string): If already managed, provide the Cinder ID.
        - extra_info (string): Any extra information to return to the user
        - source_reference (string): Similar to "reference", but for the
        snapshot's source volume.

        :param cinder_snapshots: A list of snapshots in this host that Cinder
                                 currently manages, used to determine if
                                 a snapshot is manageable or not.
        :param marker:    The last item of the previous page; we return the
                          next results after this value (after sorting)
        :param limit:     Maximum number of items to return
        :param offset:    Number of items to skip after marker
        :param sort_keys: List of keys to sort results by (valid keys are
                          'identifier' and 'size')
        :param sort_dirs: List of directions to sort by, corresponding to
                          sort_keys (valid directions are 'asc' and 'desc')

        """
        pass

    # ============
    # = Unmanage =
    # ============

    @datc.lookup
    def unmanage(self, volume):
        """Unmanage a currently managed volume in Cinder

        :param volume:       Cinder volume to unmanage
        """
        pass

    # ====================
    # = Fast Image Clone =
    # ====================

    @datc.lookup
    def clone_image(self, context, volume, image_location, image_meta,
                    image_service):
        """Clone an existing image volume."""
        pass

    # ====================
    # = Volume Migration =
    # ====================

    @datc.lookup
    def update_migrated_volume(self, context, volume, new_volume,
                               volume_status):
        """Return model update for migrated volume.

        Each driver implementing this method needs to be responsible for the
        values of _name_id and provider_location. If None is returned or either
        key is not set, it means the volume table does not need to change the
        value(s) for the key(s).
        The return format is {"_name_id": value, "provider_location": value}.
        :param volume: The original volume that was migrated to this backend
        :param new_volume: The migration volume object that was created on
                           this backend as part of the migration process
        :param original_volume_status: The status of the original volume
        :returns: model_update to update DB with any needed changes
        """
        pass

    # ================
    # = Volume Stats =
    # ================

    @datc.lookup
    def get_volume_stats(self, refresh=False):
        """Get volume stats.

        If 'refresh' is True, run update first.
        The name is a bit misleading as
        the majority of the data here is cluster
        data.
        """
        pass

    # =========
    # = Login =
    # =========

    @datc.lookup
    def login(self):
        pass

    def _get_lunid(self):
        return 0

    # ============================
    # = Volume-Types/Extra-Specs =
    # ============================

    def _init_vendor_properties(self):
        """Create a dictionary of vendor unique properties.

        This method creates a dictionary of vendor unique properties
        and returns both created dictionary and vendor name.
        Returned vendor name is used to check for name of vendor
        unique properties.

        - Vendor name shouldn't include colon(:) because of the separator
          and it is automatically replaced by underscore(_).
          ex. abc:d -> abc_d
        - Vendor prefix is equal to vendor name.
          ex. abcd
        - Vendor unique properties must start with vendor prefix + ':'.
          ex. abcd:maxIOPS

        Each backend driver needs to override this method to expose
        its own properties using _set_property() like this:

        self._set_property(
            properties,
            "vendorPrefix:specific_property",
            "Title of property",
            _("Description of property"),
            "type")

        : return dictionary of vendor unique properties
        : return vendor name

        prefix: DF --> Datera Fabric
        """
        LOG.debug("Using the following volume-type defaults: %s",
                  self.defaults)

        properties = {}

        self._set_property(
            properties,
            "DF:iops_per_gb",
            "Datera IOPS Per GB Setting",
            _("Setting this value will calculate IOPS for each volume of "
              "this type based on their size.  Eg. A setting of 100 will "
              "give a 1 GB volume 100 IOPS, but a 10 GB volume 1000 IOPS. "
              "A setting of '0' is unlimited.  This value is applied to "
              "total_iops_max and will be overridden by total_iops_max if "
              "iops_per_gb is set and a large enough volume is provisioned "
              "which would exceed total_iops_max"),
            "integer",
            minimum=0,
            default=int(self.defaults.get('iops_per_gb', 0)))

        self._set_property(
            properties,
            "DF:bandwidth_per_gb",
            "Datera Bandwidth Per GB Setting",
            _("Setting this value will calculate bandwidth for each volume of "
              "this type based on their size in KiB/s.  Eg. A setting of 100 "
              "will give a 1 GB volume 100 KiB/s bandwidth, but a 10 GB "
              "volume 1000 KiB/s bandwidth. A setting of '0' is unlimited. "
              "This value is applied to total_bandwidth_max and will be "
              "overridden by total_bandwidth_max if set and a large enough "
              "volume is provisioned which woudl exceed total_bandwidth_max"),
            "integer",
            minimum=0,
            default=int(self.defaults.get('bandwidth_per_gb', 0)))

        self._set_property(
            properties,
            "DF:placement_mode",
            "Datera Volume Placement Mode (deprecated)",
            _("'DEPRECATED: PLEASE USE 'placement_policy' on 3.3.X+ versions "
              " of the Datera product.  'single_flash' for "
              "single-flash-replica placement, "
              "'all_flash' for all-flash-replica placement, "
              "'hybrid' for hybrid placement"),
            "string",
            default=self.defaults.get('placement_mode', 'hybrid'))

        self._set_property(
            properties,
            "DF:placement_policy",
            "Datera Volume Placement Policy",
            _("Valid path to a media placement policy.  Example: "
              "/placement_policies/all-flash"),
            "string",
            default=self.defaults.get('placement_policy',
                                      'default'))

        self._set_property(
            properties,
            "DF:round_robin",
            "Datera Round Robin Portals",
            _("True to round robin the provided portals for a target"),
            "boolean",
            default="True" == self.defaults.get('round_robin', "False"))

        if self.configuration.get('datera_debug_replica_count_override'):
            replica_count = 1
        else:
            replica_count = 3
        self._set_property(
            properties,
            "DF:replica_count",
            "Datera Volume Replica Count",
            _("Specifies number of replicas for each volume. Can only be "
              "increased once volume is created"),
            "integer",
            minimum=1,
            default=int(self.defaults.get('replica_count', replica_count)))

        self._set_property(
            properties,
            "DF:ip_pool",
            "Datera IP Pool",
            _("Specifies IP pool to use for volume.  If provided string "
              "contains commas, it will be split on the commas and each "
              "substring will be uses as a separate IP pool and the volume's "
              "IP pool will be chosen randomly from the list.  Example: "
              "'my-ip-pool1,my-ip-pool2,my-ip-pool3', next attach "
              "my-ip-pool2 was chosen randomly as the volume IP pool"),
            "string",
            default=self.defaults.get('ip_pool', 'default'))

        self._set_property(
            properties,
            "DF:template",
            "Datera Template",
            _("Specifies Template to use for volume provisioning"),
            "string",
            default=self.defaults.get('template', ''))

        # ###### QoS Settings ###### #
        self._set_property(
            properties,
            "DF:read_bandwidth_max",
            "Datera QoS Max Bandwidth Read",
            _("Max read bandwidth setting for volume qos, "
              "use 0 for unlimited"),
            "integer",
            minimum=0,
            default=int(self.defaults.get('read_bandwidth_max', 0)))

        self._set_property(
            properties,
            "DF:write_bandwidth_max",
            "Datera QoS Max Bandwidth Write",
            _("Max write bandwidth setting for volume qos, "
              "use 0 for unlimited"),
            "integer",
            minimum=0,
            default=int(self.defaults.get('write_bandwidth_max', 0)))

        self._set_property(
            properties,
            "DF:total_bandwidth_max",
            "Datera QoS Max Bandwidth Total",
            _("Max total bandwidth setting for volume qos, "
              "use 0 for unlimited"),
            "integer",
            minimum=0,
            default=int(self.defaults.get('total_bandwidth_max', 0)))

        self._set_property(
            properties,
            "DF:read_iops_max",
            "Datera QoS Max iops Read",
            _("Max read iops setting for volume qos, "
              "use 0 for unlimited"),
            "integer",
            minimum=0,
            default=int(self.defaults.get('read_iops_max', 0)))

        self._set_property(
            properties,
            "DF:write_iops_max",
            "Datera QoS Max IOPS Write",
            _("Max write iops setting for volume qos, "
              "use 0 for unlimited"),
            "integer",
            minimum=0,
            default=int(self.defaults.get('write_iops_max', 0)))

        self._set_property(
            properties,
            "DF:total_iops_max",
            "Datera QoS Max IOPS Total",
            _("Max total iops setting for volume qos, "
              "use 0 for unlimited"),
            "integer",
            minimum=0,
            default=int(self.defaults.get('total_iops_max', 0)))
        # ###### End QoS Settings ###### #

        return properties, 'DF'



# Stubs for oslo_i18n.fixture (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

import fixtures
import gettext
from typing import Any, Optional

class Translation(fixtures.Fixture):
    domain: Any = ...
    def __init__(self, domain: str = ...) -> None: ...
    def lazy(self, msg: Any): ...
    def immediate(self, msg: Any): ...

class ToggleLazy(fixtures.Fixture):
    def __init__(self, enabled: Any) -> None: ...
    def setUp(self) -> None: ...

class _PrefixTranslator(gettext.NullTranslations):
    prefix: Any = ...
    def __init__(self, fp: Optional[Any] = ..., prefix: str = ...) -> None: ...
    def gettext(self, message: Any): ...
    def ugettext(self, message: Any): ...

class PrefixLazyTranslation(fixtures.Fixture):
    languages: Any = ...
    locale: Any = ...
    def __init__(self, languages: Optional[Any] = ..., locale: Optional[Any] = ...) -> None: ...
    def setUp(self): ...
